#include "mpi.h"
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include "xmit_file.h"
#include "debug.h"
#if HAVE_CONFIG_H
#	include "config.h"
#endif /* HAVE_CONFIG_H */

/* Limits */
#define BUF_SIZE			16384
#define MAX_SRC_FILES		1024
#define MAX_FILENAME_LEN 	1024
#define MAX_ERROR_LEN		1024

char *get_filename(char *path)
{
	char *filename = strrchr(path, '/');
	if (filename == NULL)
		return path;
	else return filename+1;
}

char *get_parent_dir(char *path)
{
	int len;
	char *dirname;
	char *filename;

	filename = strrchr(path, '/');
	if (filename == NULL)
		return strdup(".");

	len = filename-path+1;
	dirname = malloc(len);
	if (dirname == NULL)
		return NULL;
	strncpy(dirname, path, len-1);
	dirname[len-1] = '\0';
	return dirname;
}

static size_t num_src;
static char *src_files[MAX_SRC_FILES];
static char *dest_file;
#ifndef PTMV
static int bcast_flags = BCAST_RETRY;
#else
static int bcast_flags = BCAST_RETRY | 
                         BCAST_RECURSIVE | 
                         BCAST_PRESERVE_LINKS |
                         BCAST_PRESERVE_SPECIAL |
                         BCAST_PRESERVE_MODE;
#endif /* PTMV */
static char error[MAX_ERROR_LEN];

int parse_args(int argc, char *argv[])
{
	int i;
	int c;

	while ((c = getopt(argc, argv, "o:frRpdv")) != -1){
		switch (c) {
			case 'o':
				if (strcasecmp(optarg, "force_compression") == 0){
                    if (bcast_flags & BCAST_NO_COMPRESSION) {
                        snprintf(error, MAX_ERROR_LEN, 
                                "Both force_compression and "
                                "no_compression were specified");
                        return -1;
                    }
                    bcast_flags |= BCAST_FORCE_COMPRESSION;
				} else if (strcasecmp(optarg, "no_compression") == 0){
                    if (bcast_flags & BCAST_FORCE_COMPRESSION) {
                        snprintf(error, MAX_ERROR_LEN, 
                                "Both force_compression and "
                                "no_compression were specified");
                        return -1;
                    }
                    bcast_flags |= BCAST_NO_COMPRESSION;
                } else if (strcasecmp(optarg, "no_checking") == 0) {
                    pt_debug("No checking\n");
                    bcast_flags |= BCAST_NO_CHECKING;
                } else if (strcasecmp(optarg, "force_tar") == 0) {
                    if (bcast_flags & BCAST_NO_TAR) {
                        snprintf(error, MAX_ERROR_LEN, 
                                "Both force_tar and no_tar were specified");
                        return -1;
                    }
                    bcast_flags |= BCAST_FORCE_TAR;
                } else if (strcasecmp(optarg, "no_tar") == 0) {
                    if (bcast_flags & BCAST_FORCE_TAR) {
                        snprintf(error, MAX_ERROR_LEN, 
                                "Both force_tar and no_tar were specified");
                        return -1;
                    }
                    bcast_flags |= BCAST_NO_TAR;
                } else {
                    snprintf(error, MAX_ERROR_LEN,
                             "Unknown option -o %s", optarg);
                    return -1;
				}
				break;
            case 'f':
                bcast_flags |= BCAST_FORCE;
				break;
#ifndef PTMV
			case 'r':
                bcast_flags |= BCAST_RECURSIVE;
				break;
			case 'R':
                bcast_flags |= BCAST_RECURSIVE;
                bcast_flags |= BCAST_PRESERVE_SPECIAL;
				break;
			case 'p':
                bcast_flags |= BCAST_PRESERVE_MODE;
				break;
            case 'd':
                bcast_flags |= BCAST_PRESERVE_LINKS;
                break;
#endif /* PTMV */
            case 'v':
                bcast_flags |= BCAST_VERBOSE;
                break;
            case 'h':
			case 'i':/* We cannot support -i (interactive mode) */
			default:
				snprintf(error, MAX_ERROR_LEN, 
						 "Unsupported option '-%c'", optopt);
				return -1;
		}
	}

	/* we need at least a source and a destination */
	if (argc - optind == 0){
		snprintf(error, MAX_ERROR_LEN, "Source and destination required");
		return -1;
	}
	if (argc - optind == 1){
		snprintf(error, MAX_ERROR_LEN, "Destination required");
		return -1;
	}
	
	/* Take away one for the destination name */
	num_src = argc - optind - 1;
	pt_debug("num_src = %d\n", num_src);
	for (i = optind; i < argc - 1; i++){
		/* An argument got mixed up with the filenames */
		if (argv[i][0] == '-'){
			snprintf(error, MAX_ERROR_LEN, 
					"Option found after first source file: '%s'", argv[i]);
			return -1;
		}

		src_files[i-optind] = argv[i];
		pt_debug("src_files[%d] = '%s'\n", i-optind, argv[i]);
	}

	dest_file = argv[argc-1];
	/* An argument got mixed up with the filenames */
	if (dest_file[0] == '-'){
		snprintf(error, MAX_ERROR_LEN, 
				"Option found after first source file: '%s'", argv[i]);
		return -1;
	}
	
	pt_debug("Parsed options successfully\n");

	return 0;
}

/*D

  ptcp - SUT copy

Description:
ptcp is a utility that performs a one-to-many machine file copy.

Command Line Arguments:

ptcp [-all|-m <machine_file>|-M <machine_list>] [-rRfpd] [-o <transfer_option>] <source1> <source2>... <destination>

.N machine_args

.	-r	-
		Copy recursively, copying from pipes.  See BUGS below.

.	-R	-
		Copy recursively, preserving pipes.

.	-f	-
		Force the copy.  Remove any existing destination.

.	-p	-
		Preserve user, group, times, and permissions of the source(s).

.   -d  -
        Preserve links.

-	-o <transfer_option> -
        The manner in which ptcp transfers files can be configured by
        specifying any of the following options:

            force_compression - ptcp may attempt to use compression if 
                                it is available on the host system.  
                                Typically, it only compresses when not 
                                copying already compressed files.  Use
                                this option to force the use of compression.

            no_compression    - Use this option to disable the use of 
                                compression.

            no_checking       - Normally, ptcp does checking to ensure
                                correct data transmission.  Use this option
                                to disable this feature, possibly speeding
                                up transmission.

            force_tar         - ptcp may use tar to transfer large directories
                                to speed transfers.  Use this option to force
                                the use of tar.

            no_tar            - Use this option to disable the use of tar.

Environment Variables:

.N machine_envs

Examples:

To copy the directories /home/johndoe/project and /home/johndoe/data from the 
local machine to the directory /home/johndoe on the machines specified in the 
file "destination_machines" use

.vb
ptcp -m destination_machines -r /home/johndoe/project /home/johndoe/data /home/johndoe
.ve

To make the same copy, specifying the machines on the command line, use

.vb
ptcp -M "host1 host2 host3" -r /home/johndoe/project /home/johndoe/data /home/johndoe
.ve

To copy to similarly named hosts as in the previous example, an abbreviation
can be used

.vb
ptcp -M "host%d@1-3" -r /home/johndoe/project /home/johndoe/data /home/johndoe
.ve

where the %d is replaced by the numbers 1-3 when the copy is performed.

Bugs:

The -r option does not work correctly -- it will not read from pipes, but
simply copy them as pipes.

.seealso: ptmv, ptrm

D*/

/*D

  ptmv - SUT move

Description:
ptmv is a utility that performs a one-to-many machine file copy, then 
deletes the source file(s).

Command Line Arguments:

ptmv [-m <machine_file>|-M <machine_list>] [-f] [-o <compression_option>] <source1> <source2>... <destination>

.N machine_args

.	-f	-
		Force the copy.  Remove any existing destination.

-	-o <transfer_option> -
        The manner in which ptmv transfers files can be configured by
        specifying any of the following options:

            force_compression - ptmv may attempt to use compression if 
                                it is available on the host system.  
                                Typically, it only compresses when not 
                                copying already compressed files.  Use
                                this option to force the use of compression.

            no_compression    - Use this option to disable the use of 
                                compression.

            no_checking       - Normally, ptmv does checking to ensure
                                correct data transmission.  Use this option
                                to disable this feature, possibly speeding
                                up transmission.

            force_tar         - ptmv may use tar to transfer large directories
                                to speed transfers.  Use this option to force
                                the use of tar.

            no_tar            - Use this option to disable the use of tar.

Environment Variables:

.N machine_envs

Examples:

To move the directories /home/johndoe/project and /home/johndoe/data from the 
local machine to the directory /home/johndoe on the machines specified in the 
file "destination_machines" use

.vb
ptmv -m destination_machines /home/johndoe/project /home/johndoe/data /home/johndoe
.ve

To make the same move, specifying the machines on the command line, use

.vb
ptmv -M "host1 host2 host3" /home/johndoe/project /home/johndoe/data /home/johndoe
.ve

To copy to similarly named hosts as in the previous example, an abbreviation
can be used

.vb
ptmv -M "host%d@1-3" /home/johndoe/project /home/johndoe/data /home/johndoe
.ve

where the %d is replaced by the numbers 1-3 when the copy is performed.

.seealso: ptcp, ptrm

D*/

int main(int argc, char *argv[])
{
	int i = 0;
	int rank;
	int status;
	int global_status;
	char buf[BUF_SIZE];
	int do_delete = 1;
    int color = 1;
    struct stat st;
    MPI_Comm curr;

	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    gethostname(buf, BUF_SIZE);
    pt_debug("%d is %s\n", rank, buf);

	/* Parse the arguments */
	status = parse_args(argc, argv);
	MPI_Allreduce(&status, &global_status, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
	if (global_status < 0){
		if (rank == 0)
			fprintf(stderr, "%s\n", error);
		MPI_Finalize();
		return 1;
	}

    if (!(bcast_flags & BCAST_NO_COMPRESSION)) {
        /* Initialize the compression */
        status = compression_init();
        MPI_Allreduce(&status, &global_status, 1, MPI_INT, MPI_MIN, 
                      MPI_COMM_WORLD);
        if (global_status < 0){
            if (rank == 0)
                fprintf(stderr, "Unable to initialize compression library\n");
            MPI_Finalize();
            return 1;
        }
    }

    if (num_src > 1) {
        /* Check that the destination is a directory */
        if (rank != 0) {
            if (stat(dest_file, &st) == 0) {
                if (!S_ISDIR(st.st_mode)) {
                    color = 0;
                    snprintf(buf, BUF_SIZE, "host %d", rank);
                    gethostname(buf, BUF_SIZE);
                    fprintf(stderr, "%s is not a directory on %s\n", dest_file,
                            buf);
                } else {
                    color = 1;
                }
            }
        }

        /* Split into those whose dest is a directory and the rest */
        if (MPI_Comm_split(MPI_COMM_WORLD, color, 0, &curr) != 0) {
            if (rank == 0)
                fprintf(stderr, "Internal MPI Error\n");
            MPI_Finalize();
            return 1;
        }

        /* Quit if the dest is not a directory */
        if (color == 0) {
            MPI_Finalize();
            return 1;
        }

        /* Copy to all legit destinations */
        for (i = 0; i < num_src; i++){
            do_delete = !Bcast_file(src_files[i], dest_file, bcast_flags, 
                                    0, curr);
#ifdef PTMV
            if ((rank == 0) && (do_delete)) {
                snprintf(buf, BUF_SIZE, "rm -rf %s", src_files[i]);
                system(buf);
            }
#endif /* PTMV */
        }
    } else {
        do_delete = !Bcast_file(src_files[i], dest_file, bcast_flags, 
                                0, MPI_COMM_WORLD);
#ifdef PTMV
        if ((rank == 0) && (do_delete)) {
            snprintf(buf, BUF_SIZE, "rm -rf %s", src_files[i]);
            system(buf);
        }
#endif /* PTMV */
    }
    pt_debug("%d finalizing...\n", rank);

	MPI_Finalize();
	return 0;
}
