#include <time.h>
#include <sys/types.h>
#include <signal.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include "ptfps_procps.h"
#include "ptfps.h"

static int psdb_open = 0;

process_t *get_next_process()
{
	static proc_t **curr_proc = NULL;
	static proc_t **proctab = NULL;

    if (!psdb_open) {
        (void)open_psdb(NULL);
        psdb_open = 1;
    }
	
	if (proctab == NULL){
		proctab = readproctab(PROC_FILLMEM|PROC_FILLUSR|PROC_FILLCMD);
		if (proctab == NULL)
			return NULL;
		curr_proc = proctab;
	}

	if (*curr_proc == NULL){
		freeproctab(proctab);
        return NULL;
	}

	return *curr_proc++;
}

/* UID's */
uid_t get_ruid(process_t *p)
{
	return p->ruid;
}
uid_t get_euid(process_t *p)
{
	return p->euid;
}
uid_t get_suid(process_t *p)
{
	return p->suid;
}
uid_t get_fuid(process_t *p)
{
	return p->fuid;
}

/* GID's */
gid_t get_rgid(process_t *p)
{
	return p->rgid;
}

gid_t get_egid(process_t *p)
{
	return p->egid;
}

gid_t get_sgid(process_t *p)
{
	return p->sgid;
}

gid_t get_fgid(process_t *p)
{
	return p->fgid;
}

/* Username's */
char *get_rusername(process_t *p)
{
	return p->ruser;
}

char *get_eusername(process_t *p)
{
	return p->euser;
}

char *get_susername(process_t *p)
{
	return p->suser;
}

char *get_fusername(process_t *p)
{
	return p->fuser;
}

/* Group names */
char *get_rgroupname(process_t *p)
{
	return p->rgroup;
}

char *get_egroupname(process_t *p)
{
	return p->egroup;
}

char *get_sgroupname(process_t *p)
{
	return p->sgroup;
}

char *get_fgroupname(process_t *p)
{
	return p->fgroup;
}

/* Etc */
char *get_name(process_t *p)
{
    int i, len;
    static char buf[81];

    if (p->cmdline == NULL)
        return p->cmd;

    buf[0] = '\0';

    for (i = 0, len = 0; (p->cmdline[i] != NULL) && (len < 80); i++) {
        strncat(buf, p->cmdline[i], 80-len);
        len += strlen(p->cmdline[i]);
        if (p->cmdline[i+1] == NULL)
            break;
        strncat(buf, " ", 80-len);
        len += 1;
    }

    return buf;
}

char *get_cmd(process_t *p)
{
    return p->cmd;
}

#define MAX_TTY_LEN		128
char *get_tty(process_t *p)
{
	static char buf[MAX_TTY_LEN];

	dev_to_tty(buf, MAX_TTY_LEN, p->tty, p->pid, ABBREV_DEV);

	return buf;
}
 
status_t get_state(process_t *p)
{
    switch (p->state) {
        case 'R':
            return ST_RUNNABLE;
        case 'D':
        case 'S':
            return ST_SLEEPING;
        case 'T':
            return ST_STOPPED;
        case 'Z':
            return ST_ZOMBIE;
        default:
            return ST_UNKNOWN;
    }
}

char *get_wchan(process_t *p)
{
    char *tmp;
    static char buf[30];

    if ((p->wchan & 0xffffff) == 0)
        return "-";

    tmp = wchan(p->wchan);

    if (*tmp == '\0') {
        snprintf(buf, 30, "%lx", p->wchan & 0xffffff);
        return buf;
    }

    return tmp;
}

pid_t get_pid(process_t *p)
{
	return p->pid;
}
	
pid_t get_ppid(process_t *p)
{
	return p->ppid;
}
	
time_t get_stime(process_t *p)
{
	double utime_secs, utime_nsecs;

	uptime(&utime_secs, &utime_nsecs);

	return (time(NULL) - (time_t)utime_secs) + p->start_time/Hertz;
}

unsigned long get_time(process_t *p)
{
	return (p->utime + p->stime) / Hertz;
}

double get_pcpu(process_t *p)
{
	double t = (double)(time(NULL) - get_stime(p));
	/* Note that this is the percentage of this process' time that has been
	   spent in the cpu, not the percentage of the cpu it is currently using */
 	return t == 0.0 ? 0.0 : (double)get_time(p) / t;
}

double get_pmem(process_t *p)
{
	static unsigned long long **mem = NULL;
	static unsigned long long total_mem = 0;

	if (mem == NULL){
		mem = meminfo();
		if (mem == NULL)
			return 0;
		total_mem = mem[meminfo_main][meminfo_total];
	}
	
	return ((double)(p->resident*PAGE_SIZE)/(double)total_mem);
}

unsigned long get_rss(process_t *p)
{
    return p->resident;
}

unsigned long get_vsz(process_t *p)
{
    return p->size;
}

unsigned long get_flags(process_t *p)
{
    return p->flags & 0x777;
}

int get_pri(process_t *p)
{
    return p->priority;
}

int get_nice(process_t *p)
{
    return p->nice;
}
