#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <errno.h>
#include "debug.h"
#include "ptools.h"
#include "xmit_file.h" /*I "xmit_file.h" I*/
#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif /* HAVE_CONFIG_H */

#define PATH_SIZE   1024
#define BUF_SIZE	16384

#define NORMALIZE_RV(rv)    ((rv) < 0 ? -1 : 0)

static char buf[BUF_SIZE];
static char streambuf[BUF_SIZE];

/* Compression */

static int use_compression = 0;

#ifdef USE_LZO_COMPRESSION
#include <lzo1x.h>
static lzo_byte out[BUF_SIZE + BUF_SIZE/64 + 16 + 3];
static lzo_byte wrkmem[LZO1X_1_MEM_COMPRESS];

/*@

  compression_init - initialize the compression library

Description:

Call this function to enable the use of compression in other xmit_file calls.

Return Value:

compression_init() returns 0 on success, -1 on error.  If there is an error,
it is still possible to use xmit_file calls, but compression will not be used.

.seealso compress_buffer decompress_buffer Send_file Recv_file Send_file_fd Recv_file_fd Send_file_tar Recv_file_tar

@*/
int compression_init()
{
    if (lzo_init() == 0){
        use_compression = 1;
        return 0;
    } else {
        use_compression = 0;
        return -1;
    }
}

/*@

  compress_buffer - generic interface to compress a buffer

Description:

This is an internal function of xmit_file that is used to compress a buffer
with whatever available compression library is on the host system.  It is 
documented only for reference.

Input arguments:

+ in_buf - The buffer to be compressed
- in_len - The size of in_buf

Output arguments:

- out_buf - Pointer to the compressed buffer

Return Value:

compress_buffer() returns the size of the compressed buffer.

.seealso compression_init decompress_buffer Send_file Recv_file Send_file_fd Recv_file_fd Send_file_tar Recv_file_tar
@*/
static unsigned int compress_buffer(void *in_buf, int in_len, void **out_buf)
{
    unsigned int out_len;
    lzo1x_1_compress(in_buf, in_len, out, &out_len, wrkmem);
    if (out_len >= in_len){
        *out_buf = in_buf;
        return in_len;
    } else {
        *out_buf = out;
        return out_len;
    }
}

/*@

  decompress_buffer - generic interface to decompress a buffer

Description:

This is an internal function of xmit_file that is used to decompress a buffer
with whatever available compression library is on the host system.  It is 
documented only for reference.

Input arguments:

+ in_buf - The buffer to be decompressed
- in_len - The size of in_buf

Output arguments:

- out_buf - Pointer to the decompressed buffer

Return Value:

decompress_buffer() returns the size of the decompressed buffer.

.seealso compression_init decompress_buffer Send_file Recv_file Send_file_fd Recv_file_fd Send_file_tar Recv_file_tar
@*/
static unsigned int decompress_buffer(void *in_buf, int in_len, void **out_buf)
{
    unsigned int out_len;

    lzo1x_decompress(in_buf, in_len, out, &out_len, NULL);

    *out_buf = out;
    return out_len;
}

#elif USE_ZLIB_COMPRESSION
#include <zlib.h>

#define ZLIB_COMPRESSION_LEVEL 1

static Bytef out[BUF_SIZE + BUF_SIZE/1000 + 12];

int compression_init()
{
    use_compression = 1;

    return 0;
}

static unsigned int compress_buffer(void *in_buf, int in_len, void **out_buf)
{
    int rv;
    uLongf out_len = BUF_SIZE + BUF_SIZE/1000 + 12;

    rv = compress2(out, &out_len, in_buf, in_len, ZLIB_COMPRESSION_LEVEL);
    if ((rv != Z_OK) || (out_len >= in_len)) {
        *out_buf = in_buf;
        return in_len;
    } else {
        *out_buf = out;
        return (unsigned int)out_len;
    }
}

static unsigned int decompress_buffer(void *in_buf, int in_len, void **out_buf)
{
    uLongf out_len = BUF_SIZE + BUF_SIZE/1000 + 12;

    uncompress(out, &out_len, in_buf, in_len);

    *out_buf = out;
    return (unsigned int)out_len;
}

#else /* No compression */

int compression_init()
{
    return 0;
}
/* Set the output buffer to the input buffer, return the input size */
#define compress_buffer(a,b,c)		(*c = a, b)
#define decompress_buffer(a,b,c)	(*c = a, b)

#endif

/* Some common compression algorithm extensions */
static const char *comp_exts[] = {
    "gz", "bz2", "Z", "zip", "sea", 
    "sit", "lha", "cab", "arj", "ark", 
    "arc", "z", "lzh"};
    static const int num_comp_exts = sizeof(comp_exts)/sizeof(comp_exts[0]);
    /* Tries to guess whether a filename is compressed or not */
int is_compressed(char *fn)
{
    int i;
    char *ext = strrchr(fn, '.');

    if (ext == NULL)
        return 0;

    ext++;

    for (i = 0; i < num_comp_exts; i++){
        if (strcmp(ext, comp_exts[i]) == 0)
            return 1;
    }

    return 0;
}

/* static path helpers */

/*@

  get_last_component - gets the last component from a path

Description:

This is an internal function of xmit_file that is used to extract the filename
from a path.  get_last_component() only works based on the textual path given - 
it does not do any checking of the filesystem.

Input arguments:

- path - the path

Return Value:

get_last_component() returns the address of the beginning of the last path
component.

.seealso get_parent_dir
@*/

static char *get_last_component(char *path)
{
    char *filename = strrchr(path, '/');

    if (filename == NULL)
        return path;
    if ((filename[1] != '\0') && (filename[1] != '/'))
        return filename + 1;
    for (; filename != path; filename--) {
        if ((filename[0] == '/') && 
                (filename[1] != '/') && 
                (filename[1] != '\0'))
            return filename + 1;
    }

    return path;
}

/*@

  get_parent_dir - gets the parent directory of a path

Description:

This is an internal function of xmit_file that is used to extract the parent
directory from a path.  get_parent_dir() only works based on the textual path 
given - it does not do any checking of the filesystem.

Input arguments:

+ path - the path
- dsize - the size of buffer dest

Output arguments:

- dest - the destination buffer

Return Value:

get_parent_dir() returns a pointer to dest

.seealso get_last_component
@*/
static char *get_parent_dir(char *path, char *dest, size_t dsize)
{
    char *lc;

    strncpy(dest, path, dsize);
    lc = get_last_component(dest);
    if (lc == dest) {
        strncpy(dest, ".", dsize);
    } else {
        for (lc--; lc[0] == '/'; lc--)
            lc[0] = '\0';
    }

    return dest;
}

#include "mpi.h" /*I "mpi.h" I*/
/* Send files */
/*@

  Send_file_fd - send a file described by a file descriptor

Description:

Send_file_fd sends a file described by a file descriptor to a destination 
process that has called Recv_file_fd.

Input arguments:

+ fd - the file descriptor of the file to send
. dest - the destination process
. tag - the MPI tag with which to send the file
- comm - the MPI communicator in which to send the file 

Return Value:

Send_file_fd returns 0 on success, -1 on error.

.seealso Send_file Recv_file Recv_file_fd Send_file_tar Recv_file_tar
@*/
int Send_file_fd(int fd, int dest, int tag, MPI_Comm comm)
{
    int index;
    int count, ccount;
    MPI_Request reqs[2];
    MPI_Status mpistat;
    int ABORT_TAG = tag + 1;
    void *tmp = buf;

    /* Receive error messages */
    MPI_Irecv(NULL, 0, MPI_CHAR, dest, ABORT_TAG, comm, &reqs[1]);
    while (1){
        count = read(fd, buf, BUF_SIZE);

        if (count < 0){
            /* ABORT */
            pt_debug("Aborting send: read failed!\n");
            MPI_Send(NULL, 0, MPI_CHAR, dest, ABORT_TAG, comm);
            return XF_XMITERR;
        }

        pt_debug("read %d bytes\n", count);

        if (use_compression){
            int compressed; 

            if (count > 0)
                ccount = compress_buffer(buf, count, &tmp);
            else 
                ccount = count;
            /* This is a performance hack.  
               This keeps us from having to send "compressed" buffers that 
               are larger than the original, just to process them to get less
               data.
             */
            if (ccount < count){
                /* Send the compressed version */
                count = ccount;
                compressed = 1;
            } else {
                /* Send the original version */
                tmp = buf;
                compressed = 0;
            }

            pt_debug("The data %s compressed\n", compressed ? "is" : "is not");
            MPI_Send(&compressed, 1, MPI_INT, dest, tag, comm);
        }

        pt_debug("Sending packet of size %d\n", count);
        /* Send and wait for completion */
        MPI_Isend(tmp, count, MPI_CHAR, dest, tag, comm, &reqs[0]);
        MPI_Waitany(2, reqs, &index, &mpistat);
        if (index == 0){
            /* The send completed */
            if (count == 0)
                return XF_SUCCESS;
        } else {
            /* ABORT */
            return XF_XMITERR;
        }
    }

    return XF_SUCCESS;
}

/*@

  Send_file - send a file described by a file name

Description:

Send_file sends a file described by a file name to a destination 
process that has called Recv_file.  The file will have permissions
0644 & umask.

Input arguments:

+ filename - the filename of the file to send
. dest - the destination process
. tag - the MPI tag with which to send the file
- comm - the MPI communicator in which to send the file 

Return Value:

Send_file returns 0 on success, -1 on error.

.seealso Recv_file Send_file_fd Recv_file_fd Send_file_tar Recv_file_tar
@*/
int Send_file(char *filename, int dest, int tag, MPI_Comm comm)
{
    int fd, status;
    int ABORT_TAG = tag + 1;
    int fn_len;
    char *fn;
    MPI_Status mpistat;

    fn = get_last_component(filename);
    fd = open(filename, O_RDONLY);
    if (fd < 0){
        /* ABORT */
        MPI_Send(NULL, 0, MPI_CHAR, dest, ABORT_TAG, comm);
        return XF_FILEERR;
    }

    /* Send the filename */
    fn_len = strlen(fn) + 1;
    MPI_Send(fn, fn_len, MPI_CHAR, dest, tag, comm);
    /* Receive acknowledgement */
    MPI_Recv(NULL, 0, MPI_CHAR, dest, MPI_ANY_TAG, comm, &mpistat);
    if (mpistat.MPI_TAG != tag){
        close(fd);
        return XF_XMITERR;
    }

    status = Send_file_fd(fd, dest, tag, comm);

    close(fd);

    return status;
}

#define MAX_CMD_LEN		2048

/*@

  Send_file_tar - send a file or group of files using tar

Description:

Send_file_tar sends a file or group of files using tar to archive them to
a destination process that has called Recv_file_tar.

Input arguments:

+ filename - the name of the file or names of the files to send.  This may
be any pattern that is expandable by /bin/sh
. dest - the destination process
. tag - the MPI tag with which to send the file
- comm - the MPI communicator in which to send the file 

Return Value:

Send_file_tar returns 0 on success, -1 on error.

.seealso Send_file Recv_file Send_file_fd Recv_file_fd Send_file_tar Recv_file_tar
@*/
int Send_file_tar(char *filename, int dest, int tag, MPI_Comm comm)
{
    FILE *f;
    int fd, status;
    int ABORT_TAG = tag+1;
    int fn_len;
    char cmd[MAX_CMD_LEN];
    char dir[PATH_SIZE];
    char *fn;
    MPI_Status mpistat;

    fn = get_last_component(filename);

    get_parent_dir(filename, dir, PATH_SIZE);

    if (chdir(dir) != 0){
        MPI_Send(NULL, 0, MPI_CHAR, dest, ABORT_TAG, comm);
        return XF_FILEERR;
    } 

    fn_len = strlen(fn) + 1;
    if (fn_len + sizeof("tar cf - ") >= MAX_CMD_LEN){
        MPI_Send(NULL, 0, MPI_CHAR, dest, ABORT_TAG, comm);
        return XF_INTNERR;
    } 

    sprintf(cmd, "tar cf - %s", fn);

    f = popen(cmd, "r");
    if (f == NULL){
        MPI_Send(NULL, 0, MPI_CHAR, dest, ABORT_TAG, comm);
        return XF_FILEERR;
    }

    fd = fileno(f);

    /* Send the ready notice */
    MPI_Send(NULL, 0, MPI_CHAR, dest, tag, comm);
    /* Receive acknowledgement */
    MPI_Recv(NULL, 0, MPI_CHAR, dest, MPI_ANY_TAG, comm, &mpistat);
    if (mpistat.MPI_TAG != tag){
        pclose(f);
        return XF_XMITERR;
    }

    status = Send_file_fd(fd, dest, tag, comm);

    pclose(f);

    return status;
}

/* Recv files */
/*@

  Recv_file_fd - receive a file into a file descriptor

Description:

Recv_file_fd receives a file and writes it to the file described by a file
descriptor from a source process that has called Send_file_fd.

Input arguments:

+ fd - the file descriptor of the file to write the received file from
. source - the source process
. tag - the MPI tag with which to receive the file
- comm - the MPI communicator in which to receive the file 

Return Value:

Recv_file_fd returns 0 on success, -1 on error.

.seealso Send_file Recv_file Send_file_fd Send_file_tar Recv_file_tar
@*/

int Recv_file_fd(int fd, int source, int tag, MPI_Comm comm)
{
    int c;
    unsigned int count;
    MPI_Status mpistat;
    int ABORT_TAG = tag + 1;
    void *tmp = buf;

    while (1){
        int compressed = 0;
        if (use_compression)
            MPI_Recv(&compressed, 1, MPI_INT, source, tag, comm, &mpistat);

        MPI_Recv(buf, BUF_SIZE, MPI_CHAR, source, MPI_ANY_TAG, comm, &mpistat);

        /* ABORT? */
        if (mpistat.MPI_TAG != tag)
            return XF_XMITERR;

        /* Are we done? */
        MPI_Get_count(&mpistat, MPI_CHAR, &c);
        count = (unsigned int)c;
        pt_debug("received packet of size %d\n", c);
        if (count == 0)
            return XF_SUCCESS;

        if (compressed)
            count = decompress_buffer(buf, count, &tmp);
        else
            tmp = buf;

        /* Write out */
        if (write(fd, tmp, count) != count){
            /* ABORT */
            MPI_Send(NULL, 0, MPI_CHAR, source, ABORT_TAG, comm);
            return XF_FILEERR;
        }
    }

    return XF_SUCCESS;
}

/*@

  Recv_file - receive a file

Description:

Recv_file receives a file from a source process that has called Send_file.
It writes the filename into the supplied buffer, if enough room is available.
The file will have permissions 0644 & umask.

Input arguments:
+ fn_len - the length of the filename buffer
. source - the source process
. tag - the MPI tag with which to receive the file
- comm - the MPI communicator in which to receive the file 

Output arguments:

- filename - the buffer into which the filename will be written

Return Value:

Recv_file returns 0 on success, -1 on error.

.seealso Send_file Send_file_fd Recv_file_fd Send_file_tar Recv_file_tar
@*/
int Recv_file(char *filename, int fn_len, int source, int tag, MPI_Comm comm)
{	
    int fd, status;
    int ABORT_TAG = tag+1;
    MPI_Status mpistat;

    filename[0] = '\0';
    /* Receive the filename */
    MPI_Recv(filename, fn_len, MPI_CHAR, source, MPI_ANY_TAG, comm, &mpistat);
    if (mpistat.MPI_TAG != tag)
        return XF_XMITERR;

    /* Open the file */
    fd = open(filename, O_WRONLY|O_CREAT|O_TRUNC, 0644);
    if (fd < 0){
        /* ABORT */
        MPI_Send(NULL, 0, MPI_CHAR, source, ABORT_TAG, comm);
        return XF_FILEERR;
    }

    /* Send the acknowledgement */
    MPI_Send(NULL, 0, MPI_CHAR, source, tag, comm);

    status = Recv_file_fd(fd, source, tag, comm);

    close(fd);

    return status;
}

/*@

  Recv_file_tar - receive a file or group of files

Description:

Recv_file receives a file or group of files from a source process that has 
called Send_file_tar.

Input arguments:
+ source - the source process
. tag - the MPI tag with which to receive the file
- comm - the MPI communicator in which to receive the file 

Return Value:

Recv_file_tar returns 0 on success, -1 on error.

.seealso Send_file Recv_file Send_file_fd Recv_file_fd Send_file_tar 
@*/
int Recv_file_tar(int source, int tag, MPI_Comm comm)
{	
    FILE *f;
    int fd, status;
    int ABORT_TAG = tag+1;
    MPI_Status mpistat;

    MPI_Recv(NULL, 0, MPI_CHAR, source, MPI_ANY_TAG, comm, &mpistat);
    if (mpistat.MPI_TAG != tag)
        return XF_XMITERR;

    f = popen("tar xf -", "w");
    if (f == NULL){
        MPI_Send(NULL, 0, MPI_CHAR, source, ABORT_TAG, comm);
        return XF_FILEERR;
    }

    /* Send the acknowledgement */
    MPI_Send(NULL, 0, MPI_CHAR, source, tag, comm);

    fd = fileno(f);

    status = Recv_file_fd(fd, source, tag, comm);

    pclose(f);

    return status;
}

/*@
  collect - collect and report a status

Description:

collect() collects and reports a status.  If one or more processes failed
(i.e. status = -1), then the hostname of that process' host is reported
along with the error on stderr.

Input arguments:
+ status - the status being checked (should be 0 or -1)
. err_no - the error number
. msg - the message to be written (should contain 1 %s specifier for the hostname)
. root - the root node
- comm - the MPI communicator

Output arguments:
None

Return value:
collect returns status at non-root nodes.  Returns the sum of the the stati 
at root.

@*/
static int collect(int status, int err_no, char *msg, int root, MPI_Comm comm)
{
    int i;
    int rank;
    int rstatus = status;
    MPI_Status mpistat;
    static char hostname[PATH_SIZE] = "";

    MPI_Comm_rank(comm, &rank);

    MPI_Reduce(&status, &rstatus, 1, MPI_INT, MPI_SUM, root, comm);

    if (msg == NULL)
        return rstatus;

    if (rank == root) {
        pt_debug("rstatus = %d\n", rstatus);
        if (rstatus < 0) {
            for (i = 0; i < -rstatus; i++) {
                MPI_Recv(hostname, PATH_SIZE, MPI_CHAR, 
                        MPI_ANY_SOURCE, 0, comm, &mpistat);
                MPI_Recv(&err_no, 1, MPI_INT,
                        mpistat.MPI_SOURCE, 0, comm, &mpistat);
                snprintf(buf, PATH_SIZE, msg, 
                        hostname);
                errno = err_no;
                perror(buf);
            }
        }
        return rstatus;
    } else {
        if (status == 0)
            return 0;

        if (hostname[0] == '\0')
            gethostname(hostname, PATH_SIZE);
        MPI_Send(hostname, PATH_SIZE, MPI_CHAR, root, 0, comm);
        MPI_Send(&err_no, 1, MPI_INT, root, 0, comm);
        return -1;
    }
}

/*@
  Bcast_file_fd - Broadcast a file read from fd

Description:
Bcast_file_fd broadcasts a file read from fd

Input arguments:
+ fd - the input file descriptor at root, the output file descriptor at all other nodes
. flags - option flags (see below)
. root - the root node
- comm - the MPI communicator

Output arguments:
None

Option flags:
The following flags are available\:
+ BCAST_RETRY - If a write fails, retry upto MAX_RETRIES times to write.
. BCAST_VERBOSE - Write the status of the transfer to standard error at the root node.
- BCAST_NO_CHECKING - Do not do any status checking.  This speeds transfer but risks data integrity.

Return value:
Bcast_file_fd returns 0 on success, -1 on error

.seealso Bcast_file(3), Bcast_file_tar(3)
@*/
int Bcast_file_fd(int fd, int flags, int root, MPI_Comm comm)
{
    int compressed = 0; 
    int rank, csize;
    int written = 0, rwritten;
    int retries;
    int count, ccount = 0; 
    int err_no = 0;
    int rv;
    int print_bytes = 0;
    int special_file = 0;
    off_t fsize = 0;
    void *tmp = buf;
    struct stat st;
    char erase[81] = "";
    char line[81] = "";
    int info[2] = {0, 0};

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &csize);
    if (csize == 1)
        return 0;

    if ((rank == root) && (flags & BCAST_VERBOSE)) {
        if (fstat(fd, &st) < 0) {
            print_bytes = 1;
        } else {
            if (!S_ISREG(st.st_mode))
                special_file = 1;
            if (st.st_size == 0)
                print_bytes = 1;
        }
    }

    while (1) {
        tmp = buf;
        retries = 0;

        if (rank == root) {
            written = count = read(fd, buf, BUF_SIZE);
            err_no = errno;
            if ((use_compression) && !(flags & BCAST_NO_COMPRESSION)){
                if (count > 0)
                    ccount = compress_buffer(buf, count, &tmp);
                else 
                    ccount = count;

                /* This is a performance hack.  
                   This keeps us from having to send "compressed" buffers that 
                   are larger than the original, just to process them to get 
                   less data.
                 */
                if (ccount < count){
                    /* Send the compressed version */
                    count = ccount;
                    compressed = 1;
                } else {
                    /* Send the original version */
                    compressed = 0;
                }
                info[1] = compressed;
            } else {
                tmp = buf;
            }

            info[0] = ccount;

            if (flags & BCAST_VERBOSE) {
                /* Print a progress bar */
                fsize += count;

                if (erase[0] == '\0') {
                    erase[80] = '\0';
                    memset(erase, '\b', 80);
                }

                if (line[0] == '\0') {
                    line[0] = '[';
                    line[79] = ']';
                    line[80] = '\0';
                    memset(line + 1, ' ', 78);
                } else {
                    printf(erase);
                }

                if (print_bytes) {
                    int len = 0;

                    if (special_file)
                        len += printf("Special file: ");
                    len += printf("Bytes transmitted: %lu", fsize);
                    memset(erase, '\b', len);
                } else {
                    memset(line + 1, '.', (fsize * 78) / st.st_size);
                    printf(line);
                }

                fflush(stdout);
            }
        } else {
            written = 0;
        }

        do {
            MPI_Bcast(info, 2, MPI_INT, root, comm);
            count = info[0];
            compressed = info[1];

            if (count <= 0) {
                if (count < 0) {
                    if (rank == root) {
                        errno = err_no;
                        perror("Read error");
                    }
                    return -1;
                }

                if ((rank == root) && (flags & BCAST_VERBOSE)) {
                    printf(erase);
                    if (print_bytes) {
                        int len = 0;

                        if (special_file)
                            len += printf("Special file: ");
                        len += printf("Bytes transmitted: %lu", fsize);
                        memset(erase, '\b', len);
                    } else {
                        memset(line + 1, '.', 78);
                        printf(line);
                    }
                    fflush(stdout);
                    printf("\n");
                }

                return 0;
            }

            MPI_Bcast(tmp, count, MPI_BYTE, root, comm);

            if ((rank != root) && (written != count)) {
                if (compressed) {
                    ccount = decompress_buffer(tmp, count, &tmp);
                } else {
                    ccount = count;
                }
                written = write(fd, tmp, ccount);
                err_no = errno;
            }

            if (!(flags & BCAST_NO_CHECKING)) {
                MPI_Allreduce(&written, &rwritten, 1, MPI_INT, MPI_MIN, comm);
                pt_debug("collecting...\n");
                /* Error check */
                if ((flags & BCAST_VERBOSE) || (!(flags & BCAST_RETRY))) {
                    rv = collect(NORMALIZE_RV(written - count), err_no, 
                            "Write error on %s", root, comm);
                } else {
                    rv = collect(NORMALIZE_RV(written - count), err_no, NULL, 
                            root, comm);
                }
                pt_debug("done collecting\n", written, fd);

                retries++;
            } else {
                break;
            }

        } while ((rwritten < count) && (retries < MAX_RETRIES));

                if (retries >= MAX_RETRIES) {
                    return collect(NORMALIZE_RV(written - count), err_no, 
                            "Write error on %s", root, comm);
                }
    }

    return 0;
}

/*@
  Bcast_file - Broadcast a file

Description:
Bcast_file broadcasts a file.  The file may be a regular file, a special
file, or a directory if BCAST_RECURSIVE is given as one of the flags.

Input arguments:
+ src - the source file
. dest - the destination
. flags - option flags (see below)
. root - the root node
- comm - the MPI communicator

Output arguments:
None

Option flags:
The following flags are available\:
+ BCAST_RETRY - If a write fails, retry upto MAX_RETRIES times to write.
. BCAST_VERBOSE - Write the status of the transfer to standard error at the root node.
. BCAST_RECURSIVE - Broadcast recursively.
. BCAST_PRESERVE_LINKS - Preserve symbolic links.
. BCAST_PRESERVE_MODE - Try to preserve the mode of the file.
. BCAST_PRESERVE_SPECIAL - Preserve special files.
. BCAST_FORCE - Try to delete the destination before copying.
- BCAST_NO_CHECKING - Do not do any status checking.  This speeds transfer but risks data integrity.

Return value:
Bcast_file returns 0 on success, -1 on error

.seealso Bcast_file_fd(3), Bcast_file_tar(3)
    @*/
int Bcast_file(char *src, char *dest, int flags, int root, MPI_Comm comm)
{
    int fd;
    int rank, csize;
    int color;
    int len;
    int status, global;
    int err_no = 0;
    char rdest[PATH_SIZE];
    char buf[BUF_SIZE];
    mode_t mode;
    dev_t dev;
    MPI_Comm curr;
    struct stat st;

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &csize);

    if (!(flags & BCAST_NO_COMPRESSION) && !(flags & BCAST_FORCE_COMPRESSION)) {
        if (is_compressed(src))
            flags |= BCAST_NO_COMPRESSION;
    }

    if (rank != root) {
        src = get_last_component(src);
        if (stat(dest, &st) == 0) {
            if (S_ISDIR(st.st_mode)) {
                snprintf(rdest, PATH_SIZE, "%s/%s", dest, get_last_component(src));
                dest = rdest;
            }
        }
    }

    if (rank == root) {
        status = lstat(src, &st);
        mode = st.st_mode;
    }

    MPI_Bcast(&status, 1, MPI_INT, root, comm);
    if (status != 0) {
        if (rank == root) {
            snprintf(buf, BUF_SIZE, "Unable to stat '%s'", src);
            perror(buf);
        }

        return -1;
    }

    MPI_Bcast(&mode, sizeof(mode), MPI_BYTE, root, comm);

    if (S_ISDIR(mode) && 
        ((flags & BCAST_PRESERVE_SPECIAL) || (flags & BCAST_FORCE_TAR)) &&
        !(flags & BCAST_NO_TAR)) {
        return Bcast_file_tar(src, dest, flags, root, comm);
    }
 
    if ((flags & BCAST_PRESERVE_LINKS) && S_ISLNK(mode)) {
        if (rank == root) {
            memset(buf, 0, BUF_SIZE);
            len = readlink(src, buf, BUF_SIZE-1);
            if (errno == ENAMETOOLONG)
                len = -1;
            if (flags & BCAST_VERBOSE)
                printf("L (%s -> %s) -> %s\n", src, buf, dest);
            status = 0;
        }
        MPI_Bcast(&len, 1, MPI_INT, root, comm);
        if (len < 0) {
            if (rank == root)
                fprintf(stderr, "Unable to link '%s': Link value too long\n", src);

            return -1;
        }
        MPI_Bcast(buf, len + 1, MPI_CHAR, root, comm);
        if (rank != root)
            status = NORMALIZE_RV(symlink(buf, dest));

        return collect(status, errno, "%s failed to create link", root, comm);
    }

    if (flags & BCAST_PRESERVE_SPECIAL) {
        if (S_ISLNK(mode)) {
            /* Thus we're not preserving links... */
            if (rank == root) {
                memset(buf, 0, BUF_SIZE);
                len = readlink(src, buf, BUF_SIZE-1);
                if (errno == ENAMETOOLONG)
                    len = -1;
            }
            MPI_Bcast(&len, 1, MPI_INT, root, comm);
            if (len < 0) {
                if (rank == root) {
                    fprintf(stderr, 
                            "Unable to copy link '%s': Link value too long\n", 
                            src);
                }
                return -1;
            }
            MPI_Bcast(buf, len + 1, MPI_CHAR, root, comm);
            return Bcast_file(buf, dest, flags, root, comm);
        }
        if (S_ISCHR(mode) || S_ISBLK(mode) || S_ISFIFO(mode) || S_ISSOCK(mode)){
            if (rank == root) {
                if (flags & BCAST_VERBOSE)
                    printf("Copying special file %s\n", src);
                memcpy(&dev, &st.st_dev, sizeof(dev));
            }
            MPI_Bcast(&dev, sizeof(dev), MPI_BYTE, root, comm);

            if (rank != root)
                status = NORMALIZE_RV(mknod(dest, mode, dev));

            return collect(status, errno, "%s failed to create device", 
                    root, comm);
        }
    }

    if (S_ISDIR(mode)) {
        int rv = 0;
        int top = 1;
        DIR *current;
        DIR *stack[MAX_DIR_DEPTH];
        struct dirent *de;
        char tmp[PATH_SIZE];
        char rpath[PATH_SIZE]; /* relative path */
        char sparent[PATH_SIZE]; /* source parent */

        static MPI_Datatype detailstype;
        static int details_committed = 0;
        struct {
            enum {BF_DIR, BF_FILE, BF_DONE} type;
            mode_t mode;
            char spath[PATH_SIZE]; /* source path */
            char dpath[PATH_SIZE]; /* destination path */
        } details;

        if (!details_committed) {
            int blockcounts[3] = {1, sizeof(mode), PATH_SIZE*2};
            MPI_Datatype types[3];
            MPI_Aint displs[3];

            MPI_Address(&details.type, &displs[0]);
            MPI_Address(&details.mode, &displs[1]);
            MPI_Address(&details.spath, &displs[2]);

            types[0] = MPI_INT;
            types[1] = MPI_BYTE;
            types[2] = MPI_CHAR;

            MPI_Type_struct(3, blockcounts, displs, types, &detailstype);
            MPI_Type_commit(&detailstype);
        }

        if (!(flags & BCAST_RECURSIVE)) {
            if (rank == root) {
                fprintf(stderr, "Cannot copy directory '%s'\n", src);
                fprintf(stderr, "\t(Did you mean to specify recursive?)\n");
            }

            return -1;
        }

        if (rank == root) {
            strncpy(details.spath, src, PATH_SIZE);
            strncpy(rpath, get_last_component(src), PATH_SIZE);
            snprintf(details.dpath, PATH_SIZE, "%s/%s", dest, rpath);
            get_parent_dir(src, sparent, PATH_SIZE);
            stack[0] = opendir(src);
            if (stack[0] == NULL) {
                err_no = errno;
                status = -1;
            } else {
                status = 0;
            }

            /* Transmit whether the source directory is statable */
            MPI_Bcast(&status, 1, MPI_INT, root, comm);

            if (status < 0) {
                snprintf(buf, BUF_SIZE, "Unable to open directory '%s'", src);
                errno = err_no;
                perror(buf);
                return -1;
            }

            if (flags & BCAST_VERBOSE)
                printf("D %s -> %s\n", details.spath, details.dpath);

            /* See if anyone failed to make the directory */
            snprintf(buf, BUF_SIZE, 
                    "%%s was unable to create directory %s", src);
            collect(status, err_no, buf, root, comm);

            /* Descend the directory tree */
            while (top > 0) {
                current = stack[--top];
                while ((de = readdir(current)) != NULL) {
                    /* Skip the special directories */
                    if ((strcmp(de->d_name, "..") == 0) || 
                            (strcmp(de->d_name, ".") == 0))
                        continue;

                    /* Establish the relative, source, and destination paths */
                    len = snprintf(tmp, PATH_SIZE, "%s/%s", rpath, de->d_name);
                    if ((len < 0) || (len > PATH_SIZE)) {
                        fprintf(stderr, "Path too long: '%s/%s'\n", 
                                rpath, de->d_name);
                        continue;
                    }
                    strncpy(rpath, tmp, PATH_SIZE);

                    len = snprintf(details.spath, PATH_SIZE, 
                            "%s/%s", sparent, rpath);
                    if ((len < 0) || (len > PATH_SIZE)) {
                        fprintf(stderr, "Path too long: '%s/%s'\n", 
                                sparent, rpath);
                        continue;
                    }

                    len = snprintf(details.dpath, PATH_SIZE, 
                            "%s/%s", dest, rpath);
                    if ((len < 0) || (len > PATH_SIZE)) {
                        fprintf(stderr,"Path too long: '%s/%s'\n", dest, rpath);
                        continue;
                    }

                    /* Check the source */
                    if (lstat(details.spath, &st) < 0) {
                        err_no = errno;
                        snprintf(buf, BUF_SIZE, "Unable to stat file '%s'", 
                                details.spath);
                        errno = err_no;
                        perror(buf);
                        continue;
                    }

                    /* Recurse if this is a directory */
                    if (S_ISDIR(st.st_mode)) {
                        stack[top++] = current;
                        if (top > MAX_DIR_DEPTH) {
                            fprintf(stderr, "Maximum directory depth reached!\n");
                            break;
                        }

                        current = opendir(details.spath);
                        if (current == NULL) {
                            err_no = errno;
                            snprintf(buf, BUF_SIZE, 
                                    "Unable to open directory '%s'", src);
                            errno = err_no;
                            perror(buf);
                            current = stack[--top];
                            continue;
                        } 

                        if (flags & BCAST_VERBOSE) {
                            printf("D %s -> %s\n", 
                                    details.spath, details.dpath);
                        }

                        details.type = BF_DIR;
                        if (flags & BCAST_PRESERVE_MODE) 
                            details.mode = st.st_mode;
                        else
                            details.mode = 0755;

                        MPI_Bcast(MPI_BOTTOM, 1, detailstype, root, comm);
                        err_no = errno;

                        snprintf(buf, BUF_SIZE, 
                                "%%s was unable to create directory %s", 
                                details.dpath);
                        collect(0, err_no, buf, root, comm);
                    } else {
                        /* path is not a directory, so we get at most depth 1 
                           recursion here and depth 1 recursion from the link 
                           stuff above.  I can live with 2 levels of recursion.
                         */

                        /* Broadcast the details */
                        details.type = BF_FILE;
                        if (flags & BCAST_PRESERVE_MODE) 
                            details.mode = st.st_mode;
                        else
                            details.mode = 0755;

                        MPI_Bcast(MPI_BOTTOM, 1, detailstype, root, comm);

                        /* Send the actual file */
                        Bcast_file(details.spath, details.dpath, flags, 
                                root, comm);

                        /* See how everybody did */
                        snprintf(buf, BUF_SIZE, 
                                "%%s was unable to copy file %s", 
                                details.spath);
                        collect(0, err_no, buf, root, comm);

                        /* Move back up the tree */
                        get_parent_dir(rpath, rpath, PATH_SIZE);
                    }
                }
                get_parent_dir(rpath, rpath, PATH_SIZE);
            }
            details.type = BF_DONE;
            MPI_Bcast(MPI_BOTTOM, 1, detailstype, root, comm);
        } else {
            /* Check that the source directory was valid */
            MPI_Bcast(&status, 1, MPI_INT, root, comm);
            if (status < 0)
                return -1;

            /* Make the directory here */
            status = mkdir(rdest, mode);
            err_no = errno;

            snprintf(buf, BUF_SIZE, 
                    "%%s was unable to create directory %s", rdest);
            collect(status, err_no, buf, root, comm);

            /* Receive instructions from the root */
            while (1) {
                MPI_Bcast(MPI_BOTTOM, 1, detailstype, root, comm);

                if (details.type == BF_DONE)
                    break;

                switch (details.type) {
                    case BF_DIR:
                        rv = mkdir(details.dpath, details.mode);
                        err_no = errno;
                        snprintf(buf, BUF_SIZE, 
                                "%%s was unable to create directory %s", 
                                details.dpath);
                        collect(rv, err_no, buf, root, comm);
                        break;
                    case BF_FILE:
                        pt_debug("Receiving file '%s'\n", details.dpath);
                        rv = Bcast_file(details.spath, details.dpath, flags, 
                                root, comm);
                        err_no = errno;
                        snprintf(buf, BUF_SIZE, 
                                "%%s was unable to copy file %s", 
                                details.spath);
                        collect(rv, err_no, buf, root, comm);
                        break;
                    default:
                        continue;
                }
            }
        }

        return rv;
    }

    /* At this point, we can just open the file, read from it and send it on */

    if (rank == root) {
        if (flags & BCAST_VERBOSE)
            printf("R %s -> %s\n", src, dest);

        fd = open(src, O_RDONLY);
        err_no = errno;
    }

    MPI_Bcast(&fd, 1, MPI_INT, root, comm);
    if (fd < 0){
        if (rank == root) {
            snprintf(buf, BUF_SIZE, "Unable to open source file %s", src);
            errno = err_no;
            perror(buf);
        }
        return -1;
    }

    if (rank != root) {
        if (flags & BCAST_FORCE)
            (void)unlink(dest);
        if (lstat(dest, &st) < 0) {
            fd = open(dest, O_WRONLY|O_CREAT|O_TRUNC, mode);
        } else {
            if (S_ISDIR(st.st_mode)) {
                char *fn = get_last_component(src);
                if (fn == NULL) {
                    errno = EINVAL;
                } else {
                    snprintf(buf, BUF_SIZE, "%s/%s", dest, fn);
                    pt_debug("opening %s\n", buf);
                    fd = open(buf, O_WRONLY|O_CREAT|O_TRUNC, 0644);
                }
            } else {
                fd = open(dest, O_WRONLY|O_CREAT|O_TRUNC, 0644);
            }
        }
    }

    err_no = errno;
    snprintf(buf, BUF_SIZE, "%%s was unable to open %s", dest);

    collect(NORMALIZE_RV(fd), err_no, buf, root, comm);
    MPI_Allreduce(&status, &global, 1, MPI_INT, MPI_MIN, comm);
    if (global < 0) {
        if (fd < 0)
            color = 0;
        else 
            color = 1;

        if (MPI_Comm_split(comm, color, rank, &curr) != 0){
            if (fd >= 0)
                close(fd);
            if (rank == root)
                fprintf(stderr, "Internal MPI error\n");
            return -1;
        }
        if (color)
            status = Bcast_file_fd(fd, flags, root, curr);
    } else {
        status = Bcast_file_fd(fd, flags, root, comm);
    }

    if ((rank != root) && (fd >= 0) && (flags & BCAST_PRESERVE_MODE)) {
        (void)fchmod(fd, mode);
    }

    close(fd);

    pt_debug("status = %d\n", status);

    return status;
}

/*@
  Bcast_file_tar - Broadcast a file using tar

Description:
Bcast_file_tar broadcasts a file using tar to archive the file first.  The 
file may be a regular file, a special file, or a directory if BCAST_RECURSIVE 
is given as one of the flags.

Input arguments:
+ src - the source file
. dest - the destination
. flags - option flags (see below)
. root - the root node
- comm - the MPI communicator

Output arguments:
None

Option flags:
The following flags are available\:
+ BCAST_RETRY - If a write fails, retry upto MAX_RETRIES times to write.
. BCAST_VERBOSE - Write the status of the transfer to standard error at the root node.
. BCAST_RECURSIVE - Broadcast recursively.
. BCAST_FORCE - Try to delete the destination before copying.
- BCAST_NO_CHECKING - Do not do any status checking.  This speeds transfer but risks data integrity.

Return value:
Bcast_file_tar returns 0 on success, -1 on error

.seealso Bcast_file_fd(3), Bcast_file(3)
@*/
int Bcast_file_tar(char *src, char *dest, int flags, int root, MPI_Comm comm)
{
    int fd = -1;
    int global;
    int rank, csize;
    int color;
    int do_rename = 0;
    int status = 0;
    int err_no = 0;
    char cmd[MAX_CMD_LEN];
    char dir[PATH_SIZE];
    char *fn;
    FILE *f = NULL;
    MPI_Comm curr;

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &csize);

    /* Tar up the source */
    if (rank == root){
        int len;

        fn = get_last_component(src);
        get_parent_dir(src, dir, PATH_SIZE);

        if (flags & BCAST_PRESERVE_LINKS)
            len = snprintf(cmd, MAX_CMD_LEN, "tar %ccfC - %s %s", 
                           (flags & BCAST_VERBOSE) ? 'v' : ' ', dir, fn);
        else
            len = snprintf(cmd, MAX_CMD_LEN, "tar %ccfhC - %s %s",
                           (flags & BCAST_VERBOSE) ? 'v' : ' ', dir, fn);

        if ((len < 0) || (len > MAX_CMD_LEN)) {
            fprintf(stderr, "Bcast_file_tar: Filename too long: '%s'\n", src);
            status = -1;
        } else {
            f = popen(cmd, "r");
            if (f == NULL) {
                fprintf(stderr, "Bcast_file_tar: Unable to execute tar\n");
                status = -1;
            } else {
                struct stat st;

                /* Get a buffer that's the same size as ours */
                setvbuf(f, streambuf, _IOFBF, BUF_SIZE);
                fd = fileno(f);
                if (fstat(fd, &st) == 0) {
                    if (!S_ISDIR(st.st_mode))
                        do_rename = 1;
                }
            }
        }
    }

    MPI_Bcast(&status, 1, MPI_INT, root, comm);
    if (status < 0) {
        fprintf(stderr, "Bcast_file_tar: Unable to open %s\n", src);
        return -1;
    }

    MPI_Bcast(&do_rename, 1, MPI_INT, root, comm);

    /* Start opening the destinations */
    if (rank != root){
        get_parent_dir(dest, dir, PATH_SIZE);
        if (chdir(dir) == 0) {
            f = popen("tar xf -", "w");
            if (f == NULL) {
                err_no = errno;
                status = -1;
            } else {
                /* Get a buffer that's the same size as ours */
                setvbuf(f, streambuf, _IOFBF, BUF_SIZE);
                fd = fileno(f);
            }
        } else {
            status = -1;
        }
    }

    snprintf(buf, BUF_SIZE, "%%s was unable to open %s", dest);
    collect(status, err_no, buf, root, comm);
    MPI_Allreduce(&status, &global, 1, MPI_INT, MPI_MIN, comm);
    if (global < 0){
        if (fd < 0)
            color = 0;
        else 
            color = 1;

        if (MPI_Comm_split(comm, color, rank, &curr) != 0) {
            if (fd >= 0)
                pclose(f);
            pt_debug("Bcast_file: MPI_Comm_split failed\n");
            return -1;
        }

        if (color == 1)
            status = Bcast_file_fd(fd, flags, root, curr);

        MPI_Comm_free(&curr);
    } else {
        status = Bcast_file_fd(fd, flags, root, comm);
    }

    if ((rank != root) && (do_rename) && (status == 0)) {
        status = rename(get_last_component(dest), get_last_component(src));
        if (status < 0) {
            if (gethostname(buf, BUF_SIZE) != 0)
                snprintf(buf, BUF_SIZE, "Host %d was unable to rename %s", rank, dest);
            else
                snprintf(buf, BUF_SIZE, "%s was unable to rename %s", buf, dest);
            perror(buf);
        }
    }

    if (f != NULL)
        pclose(f);

    return status;
}
